#ifndef BENCHMARKS_0FEA723A_C86F_6753_04ABD475F6FCA5FB
#define BENCHMARKS_0FEA723A_C86F_6753_04ABD475F6FCA5FB

#include "config.hpp"

namespace sokoengine {

int LIBSOKOENGINE_API run_benchmarks(double pivot_speed = 0);

} // namespace sokoengine

#endif // HEADER_GUARD
/// @file
